import * as lambda from "@aws-cdk/aws-lambda";
import * as cdk from "@aws-cdk/core";
/**
 * Properties for a NodejsFunction
 */
export interface NodejsFunctionProps extends lambda.FunctionOptions {
    /**
     * Path to the entry file (JavaScript or TypeScript), relative to your project root
     */
    readonly entry: string;
    /**
     * The name of the exported handler in the entry file.
     *
     * @default "handler"
     */
    readonly handler?: string;
    /**
     * The runtime environment. Only runtimes of the Node.js family are
     * supported.
     *
     * @default - `NODEJS_14_X` if `process.versions.node` >= '14.0.0',
     * `NODEJS_12_X` otherwise.
     */
    readonly runtime?: lambda.Runtime;
    /**
     * If you get "Module not found: Error: Can't resolve 'module_name'" errors, and you're not
     * actually using those modules, then it means there's a module you're using that is trying to
     * dynamically require other modules. This is the case with Knex.js. When this happens, pass all the modules
     * names found in the build error in this array.
     *
     * Example if you're only using PostgreSQL with Knex.js, use:
     *  `modulesToIgnore: ["mssql", "pg-native", "pg-query-stream", "tedious"]`
     */
    readonly modulesToIgnore?: string[];
    /**
     * Externals not to be bundled with your lambda, default to all Node.js builtin modules and aws-sdk. Modules I use this for: @sentry/serverless for example
     */
    readonly externals?: ["aws-sdk"];
    /**
     * Whether to automatically reuse TCP connections when working with the AWS
     * SDK for JavaScript.
     *
     * This sets the `AWS_NODEJS_CONNECTION_REUSE_ENABLED` environment variable
     * to `1`.
     *
     * @see https://docs.aws.amazon.com/sdk-for-javascript/v2/developer-guide/node-reusing-connections.html
     *
     * @default true
     */
    readonly awsSdkConnectionReuse?: boolean;
}
/**
 * A Node.js Lambda function bundled using Parcel
 */
export declare class NodejsFunction extends lambda.Function {
    constructor(scope: cdk.Construct, id: string, props?: NodejsFunctionProps);
}
