import { existsSync, mkdtempSync, writeFileSync } from 'fs';
import { tmpdir } from 'os';
import { resolve, join } from 'path';
import { versions, cwd, exit } from 'process';
import { Function, RuntimeFamily, Runtime, Code } from '@aws-cdk/aws-lambda';
import { sync } from 'cross-spawn';
import findUp from 'find-up';

/**
 * A Node.js Lambda function bundled using Parcel
 */

class NodejsFunction extends Function {
  constructor(scope, id, props = {
    entry: ""
  }) {
    var _props$externals, _props$handler, _props$runtime, _props$awsSdkConnecti;

    if (props.runtime && props.runtime.family !== RuntimeFamily.NODEJS) {
      throw new Error("Only `NODEJS` runtimes are supported.");
    }

    if (!/\.(js|ts)$/.test(props.entry)) {
      throw new Error("Only JavaScript or TypeScript entry files are supported.");
    }

    const entryFullPath = resolve(props.entry);

    if (!existsSync(entryFullPath)) {
      throw new Error(`Cannot find entry file at ${entryFullPath}`);
    }

    const userExternals = (_props$externals = props.externals) != null ? _props$externals : [];
    const defaultExternals = ["aws-sdk"];
    const handler = (_props$handler = props.handler) != null ? _props$handler : "handler";
    const defaultRunTime = nodeMajorVersion() >= 14 ? Runtime.NODEJS_14_X : Runtime.NODEJS_12_X;
    const runtime = (_props$runtime = props.runtime) != null ? _props$runtime : defaultRunTime;
    const outputDir = mkdtempSync(join(tmpdir(), "aws-lambda-nodejs-webpack"));
    const webpackConfigPath = join(outputDir, "webpack.config.js");

    const webpackBinPath = require.resolve("webpack-cli/bin/cli.js", {
      paths: [__dirname]
    });

    const pluginsPath = join(webpackBinPath.slice(0, webpackBinPath.lastIndexOf("/node_modules")), "node_modules");
    const pluginsPaths = {
      webpack: findModulePath("webpack", pluginsPath),
      "babel-loader": findModulePath("babel-loader", pluginsPath),
      "@babel/preset-env": findModulePath("@babel/preset-env", pluginsPath),
      "@babel/plugin-transform-runtime": findModulePath("@babel/plugin-transform-runtime", pluginsPath),
      "babel-plugin-source-map-support": findModulePath("babel-plugin-source-map-support", pluginsPath),
      noop2: findModulePath("noop2", pluginsPath),
      "terser-webpack-plugin": findModulePath("terser-webpack-plugin", pluginsPath)
    }; // NodeJs reserves '\' as an escape char; but pluginsPaths etc are inlined directly in the
    // TemplateString below, so will contain this escape character on paths computed when running
    // the Construct on a Windows machine, and so we need to escape these chars before writing them

    const escapePathForNodeJs = path => {
      return path.replace(/\\/g, "\\\\");
    };

    const webpackConfiguration = `
    const { builtinModules } = require("module");
    const { NormalModuleReplacementPlugin } = require("${escapePathForNodeJs(pluginsPaths["webpack"])}");
    const TerserPlugin = require("${escapePathForNodeJs(pluginsPaths["terser-webpack-plugin"])}");

    module.exports = {
      name: "aws-lambda-nodejs-webpack",
      mode: "none",
      entry: "${escapePathForNodeJs(entryFullPath)}",
      target: "node",
      resolve: {
        // next line allows resolving not found modules to local versions (require("lib/log"))
        modules: ["node_modules", "${escapePathForNodeJs(cwd())}"],
        extensions: [ ".ts", ".js" ],
      },
      context: "${escapePathForNodeJs(cwd())}",
      devtool: "source-map",
      module: {
        rules: [
          {
            test: /\\.js$/,
            exclude: /node_modules/,
            use: {
              loader: "${escapePathForNodeJs(pluginsPaths["babel-loader"])}",
              options: {
                babelrc: false, // do not use babelrc when present (could pollute lambda configuration)
                cwd: "${escapePathForNodeJs(cwd())}",
                cacheDirectory: "${escapePathForNodeJs(join(cwd(), "node_modules", ".cache", "aws-lambda-nodejs-webpack", "babel"))}",
                presets: [
                  [
                    "${escapePathForNodeJs(pluginsPaths["@babel/preset-env"])}",
                    {
                      "targets": {
                        "node": "${runtime.name.split("nodejs")[1].split(".")[0]}"
                      },
                      loose: true,
                      bugfixes: true,
                      ignoreBrowserslistConfig: true // do not use browser list configuration, we build for node X that's it
                    },
                  ]
                ],
                plugins: [
                  "${escapePathForNodeJs(pluginsPaths["@babel/plugin-transform-runtime"])}",
                  "${escapePathForNodeJs(pluginsPaths["babel-plugin-source-map-support"])}"
                ]
              }
            }
          },
          {
            test: /\\.ts$/,
            use: {
              loader: "${escapePathForNodeJs(findModulePath("ts-loader", pluginsPath))}",
              options: {
                context: "${escapePathForNodeJs(cwd())}",
                configFile: "${escapePathForNodeJs(join(cwd(), "tsconfig.json"))}",
                transpileOnly: true,
                // from: https://www.npmjs.com/package/@tsconfig/node14
                compilerOptions: {
                  lib: ["es2020"],
                  module: "commonjs",
                  target: "es2020",
                  baseUrl: ".",
                  strict: true,
                  esModuleInterop: true,
                  skipLibCheck: true,
                  forceConsistentCasingInFileNames: true
                }
              }
            },
            exclude: /node_modules/,
          },
        ]
      },
      cache: {
        type: "filesystem",
        buildDependencies: {
          config: [__filename, "${escapePathForNodeJs(join(cwd(), "tsconfig.json"))}"]
        },
        cacheDirectory: "${escapePathForNodeJs(join(cwd(), "node_modules", ".cache", "aws-lambda-nodejs-webpack", "webpack"))}"
      },
      // note: we specifically do not minify our code for Node.js
      // I have had horrible experience with code being minified for ES5 that would break on Node 12
      // If you have a good minifier that can minify to target Node 12 then open a PR
      optimization: {
        splitChunks: {
          cacheGroups: {
            vendor: {
              chunks: "all",
              filename: "vendor.js", // put all node_modules into vendor.js
              name: "vendor",
              test: /node_modules/,
            },
          },
        },
      },
      externals: [...builtinModules, ...${JSON.stringify([...defaultExternals, ...userExternals])}],
      output: {
        filename: "[name].js",
        path: "${escapePathForNodeJs(outputDir)}",
        libraryTarget: "commonjs2",
      },
      ${props.modulesToIgnore && `
      plugins: [
        new NormalModuleReplacementPlugin(
          /${escapePathForNodeJs(props.modulesToIgnore.join("|"))}/,
          "${escapePathForNodeJs(pluginsPaths["noop2"])}",
        ),
      ]
      ` || ""}
    };`;
    writeFileSync(webpackConfigPath, webpackConfiguration);
    console.time(`aws-lambda-nodejs-webpack-${props.entry}`);
    const webpack = sync(webpackBinPath, ["--config", webpackConfigPath], {
      // we force CWD to the output dir to avoid being "polluted" by any babelrc or other configuration files
      // that could mess up with our own webpack configuration. If you need to reuse your babelrc then please open an issue
      cwd: outputDir
    });
    console.timeEnd(`aws-lambda-nodejs-webpack-${props.entry}`);

    if (webpack.status !== 0) {
      var _webpack$output;

      console.error(`webpack had an error when bundling. Return status was ${webpack.status}`);
      console.error(webpack.error);
      console.error(webpack == null ? void 0 : (_webpack$output = webpack.output) == null ? void 0 : _webpack$output.map(out => {
        return out == null ? void 0 : out.toString();
      }));
      console.error("webpack configuration was:", webpackConfiguration);
      exit(1);
    }

    super(scope, id, { ...props,
      runtime,
      code: Code.fromAsset(outputDir),
      handler: `main.${handler}`
    }); // Enable connection reuse for aws-sdk

    if ((_props$awsSdkConnecti = props.awsSdkConnectionReuse) != null ? _props$awsSdkConnecti : true) {
      this.addEnvironment("AWS_NODEJS_CONNECTION_REUSE_ENABLED", "1");
    }

    this.addEnvironment("NODE_OPTIONS", "--enable-source-maps");
  }

}

function nodeMajorVersion() {
  return parseInt(versions.node.split(".")[0], 10);
} // this method forces resolving plugins relative to node_modules/aws-lambda-nodejs-webpack
// otherwise they would be resolved to the user versions / undefined versions


function findModulePath(moduleName, pluginsPath) {
  try {
    return require.resolve(moduleName, {
      paths: [__dirname]
    });
  } catch (error) {
    const modulePath = findUp.sync(moduleName, {
      type: "directory",
      cwd: pluginsPath
    });

    if (modulePath === undefined) {
      throw new Error(`Can't find module named ${moduleName} via require.resolve or find-up`);
    }

    return modulePath;
  }
}

export { NodejsFunction };
//# sourceMappingURL=aws-lambda-nodejs-webpack.esm.js.map
